"""Setup Logging"""
import logging
import sys
import os

APP_LOGGER_NAME = "aws-control-tower-manifest-builder"
LOGLEVEL = os.environ.get("LOGLEVEL", "INFO").upper()


def setup_applevel_logger(logger_name=APP_LOGGER_NAME, file_name=None):
    """
    Create and Setup logger

    Parameters:
    logger_name(string): Name of the logger
    file_name(string): Location to output logs too

    Return:
    logger from logging
    """
    logger = logging.getLogger(logger_name)
    level = logging.getLevelName(LOGLEVEL)
    logger.setLevel(level)
    formatter = logging.Formatter(
        "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
    )
    stream_handler = logging.StreamHandler(sys.stdout)
    stream_handler.setFormatter(formatter)
    logger.handlers.clear()
    logger.addHandler(stream_handler)
    if file_name:
        file_handler = logging.FileHandler(file_name)
        file_handler.setFormatter(formatter)
        logger.addHandler(file_handler)
    return logger


def get_logger(module_name):
    """Get logger for a given module name"""
    return logging.getLogger(APP_LOGGER_NAME).getChild(module_name)



"""Main module."""

import os
import sys
from datetime import date
from jinja2 import Environment, FileSystemLoader
from prettytable import PrettyTable
from . import manifest_input
from . import logger

log = logger.setup_applevel_logger()
SPACING = "\n\t\t\t\t\t\t\t\t\t"
JINJA_MANIFEST_PATH = os.path.dirname(__file__)


def main(args: None):
    """
    Process user arguments and render Control Tower manifest

    Parameters:
    args(dictionary): dicitonary of arguments
    {"input_cf": PATH}
    """

    if args:
        output_manifest_file = os.path.join(args.output, "manifest.yaml")
        default_region = "us-east-1"

    final_report = PrettyTable(["Type", "Successes", "Failures", "Totals"])

    log.info("Processing CF Templates from %s", args.input_cf)
    cf_resources, cf_successes, cf_failures = loop_through_files(
        args.input_cf, manifest_input.CfTemplate, default_region
    )
    final_report.add_row(
        [
            manifest_input.CfTemplate.__name__,
            cf_successes,
            cf_failures,
            cf_successes + cf_failures,
        ]
    )

    log.info("Processing SCPs from %s", args.input_scp)
    scp_resources, scp_successes, scp_failures = loop_through_files(
        args.input_scp, manifest_input.Scp, default_region
    )
    final_report.add_row(
        [
            manifest_input.Scp.__name__,
            scp_successes,
            scp_failures,
            scp_successes + scp_failures,
        ]
    )
    final_report.add_row(["-------", "-------", "-------", "-------"])
    final_report.add_row(
        [
            "Totals",
            cf_successes + scp_successes,
            cf_failures + scp_failures,
            cf_successes + scp_successes + cf_failures + scp_failures,
        ]
    )

    data = {"resources": cf_resources + scp_resources}
    data["region"] = default_region
    data["date"] = date.today().strftime("%y-%m-%d")

    loaded_environment = Environment(loader=FileSystemLoader(JINJA_MANIFEST_PATH))
    manifest = loaded_environment.get_template("manifest.yaml.j2").render(data)
    manifest_dict = manifest_input.ManifestInput.load_yaml(manifest, False)

    log.info(
        "Results from procesing \n CF templates %s \n SCPs %s",
        args.input_cf,
        args.input_scp,
    )
    log.info(final_report)

    if args.abort_if_error and (cf_failures + scp_failures) > 0:
        log.info("Skiping writing output YAML file due to failures")
        sys.exit(0)
    log.info("Writing output YAML file %s", output_manifest_file)
    manifest_input.ManifestInput.write_yaml(output_manifest_file, manifest_dict)


def loop_through_files(
    path: str, manifest_type: manifest_input, default_region: str
) -> list:
    """
    Loop through path and create list of either CfTemplate resources or SCPs

    Parameters:
    path(str): location of CFN or SCP
    manifest_type(manifest_input): which current manifest_type [SCP or CFN] thats beeing looked for.

    Return:
    [resources(list), successes(int), failures(int)]
    """
    resources = []
    successes = 0
    failures = 0
    for filename in sorted(os.listdir(path)):
        path_to_file = os.path.join(path, filename)
        if os.path.isfile(path_to_file) and ".yaml" in filename:
            new_input = manifest_type(path_to_file, default_region)
            if new_input.metadata_dict and not new_input.error:
                log.info("Processed .. %s", path_to_file)
                resources.append(new_input.metadata_dict)
                successes += 1
            else:
                log.info(
                    "Failed to Process .. %s, %sError -> %s",
                    path_to_file,
                    SPACING,
                    new_input.error,
                )
                failures += 1
    return resources, successes, failures


if __name__ == "__main__":
    sys.exit(main(None))  # pragma: no cover



"""Argparse CLI Functions"""

import sys
import os
import argparse
from aws_control_tower_manifest_builder import aws_control_tower_manifest_builder
from aws_control_tower_manifest_builder import logger

log = logger.get_logger(__name__)

# TODO: Add verification for inputs
# TODO: Fix Tox. Add
# TODO: adding final schema validation
# TODO: Read me and Docs. connect github and read the docs account.


def dir_path(string):
    """
    Determine if string is a path

    Parameter:
    string(string): string representation of a path

    Return:
    string(string): string representation of a path
    """
    if os.path.isdir(string):
        return string
    raise NotADirectoryError(f"Directory '{string}' does not exist")


def str2bool(input_bool):
    """
    Determine if input is True or False

    Parameter:
    string(string): string representation of a bool

    Return:
    string(string): string representation of a bool
    """
    if isinstance(input_bool, bool):
        return input_bool
    if input_bool.lower() == "true":
        return True
    if input_bool.lower() == "false":
        return False
    raise argparse.ArgumentTypeError(
        f"Boolean value expected. Received -> {input_bool}"
    )


def main():
    """Console script for aws_control_tower_manifest_builder."""
    parser = argparse.ArgumentParser(
        description="Produces the manifest.yaml file that works as input \
            for AWS Control Tower"
    )

    parser.add_argument(
        "--abort-if-error",
        "-a",
        default=False,
        type=str2bool,
        help="If set, does not produce the manifest file if any of the input \
            files could not be processed",
    )
    parser.add_argument(
        "--default-region",
        "-r",
        type=str,
        help="Default region for templates with no regio. Default us-east-1",
        default="us-east-1",
    )

    parser.add_argument(
        "--input-cf",
        "-c",
        metavar="/path/",
        type=dir_path,
        required=True,
        help="the path to the directory containing the cloud formation input \
            files",
    )

    parser.add_argument(
        "--input-scp",
        "-s",
        metavar="/path/",
        type=dir_path,
        required=True,
        help="the path to the directory containing the service control policy \
            input files",
    )

    parser.add_argument(
        "--output",
        "-o",
        metavar="/path/",
        type=dir_path,
        help="the path to store the output manifest.yaml file",
    )

    try:
        args = parser.parse_args()
    except NotADirectoryError as error:
        log.error(error)
        sys.exit()
    except argparse.ArgumentTypeError as error:
        log.error(error)
        sys.exit()

    aws_control_tower_manifest_builder.main(args)


if __name__ == "__main__":
    sys.exit(main())  # pragma: no cover



"""Tests for `aws_control_tower_manifest_builder.aws_control_tower_manifest_builder` package."""

import pytest
from src.aws_control_tower_manifest_builder.aws_control_tower_manifest_builder import (
    loop_through_files,
)
from src.aws_control_tower_manifest_builder.manifest_input import Scp, CfTemplate

DEFAULT_REGION = "us-east-1"


OUTPUT_CF = (
    [
        {
            "name": "testS3",
            "resource_file": "https://s3.amazonaws.com/solutions-reference/\
customizations-for-aws-control-tower/latest/custom-control-tower-initiation.template",
            "accounts": ["123456789012", "987456123989"],
            "organizational_units": ["dev", "prod"],
            "regions": ["us-east-1", "us-east-2"],
            "deploy_method": "stack_set",
        },
        {
            "name": "detailed_template",
            "deploy_method": "stack_set",
            "accounts": ["123456789012", "987456123989"],
            "organizational_units": ["dev", "prod"],
            "regions": ["us-east-1", "us-east-2"],
            "parameters": [
                {"parameter_key": "parameter1", "parameter_value": "value1"},
                {"parameter_key": "parameter2", "parameter_value": "value2"},
            ],
            "export_outputs": [
                {
                    "name": "/org/member/test-ssm/app-id",
                    "value": "$[output_ApplicationId]",
                }
            ],
            "resource_file": "tests/sample_templates/cf_template_detailed.yaml",
        },
        {
            "name": "templateLocalIAM",
            "description": "Template to deploy baseline IAM resources",
            "accounts": ["123456789012", "987456123989"],
            "organizational_units": ["dev", "prod"],
            "regions": ["us-east-1", "us-east-2"],
            "resource_file": "tests/sample_templates/cf_template_local.yaml",
            "deploy_method": "stack_set",
        },
        {
            "accounts": ["123456789012", "987456123989"],
            "name": "cf_template_minimal",
            "region": "us-east-1",
            "resource_file": "tests/sample_templates/cf_template_minimal.yaml",
            "deploy_method": "stack_set",
        },
    ],
    4,
    4,
)

OUTPUT_SCP = (
    [
        {
            "name": "ec2_deny",
            "accounts": ["123456789012", "987456123989"],
            "organizational_units": ["dev", "prod"],
            "regions": ["us-east-1", "us-east-2"],
            "resource_file": "tests/sample_scp/ec2_deny.yaml",
            "deploy_method": "scp",
        }
    ],
    1,
    1,
)

input_data = [
    pytest.param("tests/sample_scp", Scp, DEFAULT_REGION, OUTPUT_SCP),
    pytest.param("tests/sample_templates", CfTemplate, DEFAULT_REGION, OUTPUT_CF),
]


@pytest.mark.parametrize("path, manifest_type, default_region, output", input_data)
def test_loop_through_files_success(path, manifest_type, default_region, output):
    """Test that number of successes match"""
    response = loop_through_files(path, manifest_type, default_region)
    assert response[1] == output[1]


@pytest.mark.parametrize("path, manifest_type, default_region, output", input_data)
def test_loop_through_files_failures(path, manifest_type, default_region, output):
    """Test that number of failures match"""
    response = loop_through_files(path, manifest_type, default_region)
    assert response[2] == output[2]


@pytest.mark.parametrize("path, manifest_type, default_region, output", input_data)
def test_loop_through_files_resources(path, manifest_type, default_region, output):
    """Test that output matches"""
    response = loop_through_files(path, manifest_type, default_region)
    assert response[0] == output[0]



"""Class for parsing and structuring YAML and JSON"""
import os
import re
import ruamel.yaml
from aws_control_tower_manifest_builder import logger

log = logger.get_logger(__name__)


class ManifestInput:
    """Base Class for SCP and Manifest Objects"""

    def __init__(self, filename, region):
        """
        Base Object to structure Manifest and SCP objects

        Parameters:
        filename(string): Name of file to be loaded
        region(string of aws region): aws region
        """
        self.filename = filename
        file_dict = self.load_yaml(self.filename, True)
        try:
            file_dict["Metadata"]["manifest_parameters"]
        except KeyError:
            self.error = "File does not contain the required metadata"
            return {}, self.error
        self.metadata_dict = (
            self.load_yaml(self.filename, True)
            .get("Metadata")
            .get("manifest_parameters")
        )
        self.name = os.path.basename(filename).split(".")[0]
        self.default_region = region
        if "name" not in self.metadata_dict.keys():
            self.metadata_dict["name"] = self.name
        if "regions" not in self.metadata_dict.keys():
            self.metadata_dict["region"] = self.default_region
        if (
            "accounts" not in self.metadata_dict.keys()
            and "organizational_units" not in self.metadata_dict.keys()
        ):
            self.error = "Missing OU or accounts"
            return {}, self.error
        if "accounts" in self.metadata_dict.keys():
            for account in self.metadata_dict.get("accounts"):
                if not isinstance(account, str):
                    self.error = "Account provided is not a string"
                elif not re.match("[0-9]{12}", account):
                    self.error = "Account provided is not 12 digit"
        if "resource_file" not in self.metadata_dict.keys():
            self.metadata_dict["resource_file"] = self.filename

    @staticmethod
    def load_yaml(content: str, is_file: bool) -> dict:
        """Loads YAML from either a file or a String

        returns: dict
        """
        yaml_dict = {}
        yaml = ruamel.yaml.YAML()
        yaml.default_flow_style = False
        yaml.preserve_quotes = True
        try:
            if is_file:
                with open(content, encoding="utf-8") as yaml_file:
                    yaml_dict = yaml.load(yaml_file.read())
            else:
                yaml_dict = yaml.load(content)
        except (IOError, ruamel.yaml.YAMLError) as err:
            log.error(f"Unable to open {content} - {err}")
            return yaml_dict
        if yaml_dict is None:
            log.error("Error, yaml is empty")
            yaml_dict = {}
        return yaml_dict

    @staticmethod
    def write_yaml(filename: str, content: dict):
        """
        Writes yaml file
        """
        yaml = ruamel.yaml.YAML()
        yaml.default_flow_style = False
        try:
            with open(filename, "w", encoding="utf-8") as output_file:
                yaml.dump(content, output_file)
            output_file.close()
        except IOError as err:
            log.error(f"Unable to open {filename} - {err}")


class CfTemplate(ManifestInput):
    """
    Object for structuring CFN Templates
    """

    def __init__(self, filename, region):
        """
        Object for structuring CFN Templates
        """
        self.error = ""
        self.metadata_dict = {}
        self.deploy_method = "stack_set"
        super().__init__(filename, region)
        self.metadata_dict["deploy_method"] = self.deploy_method


class Scp(ManifestInput):
    """
    Object for structuring CFN Templates
    """

    def __init__(self, filename, region):
        """
        Object for structuring CFN Templates
        """
        self.error = ""
        self.deploy_method = "scp"
        self.metadata_dict = {}
        file = filename.replace("yaml", "json")
        if not os.path.exists(filename.replace("yaml", "json")):
            print(f"lookign for {file} if exist ")
            self.error = "File does not have corresponding json file"
        super().__init__(filename, region)
        self.filename = filename.replace("yaml", "json")
        self.metadata_dict["deploy_method"] = self.deploy_method
